from opentrons.protocol_api import ProtocolContext
from typing import Callable, Optional
from threading import Thread
import cherrypy
import datetime
import copy
import json
import time
import os
import ipaddress


DEFAULT_REST_KWARGS = dict(
    favicon_url="https://opentrons.com/icons/icon-48x48.png",
    config={
        "global": {
            "server.socket_host": "::",
            "server.socket_port": 8080,
            "engine.autoreload.on": False,
        },
        "/favicon.ico": {
            "tools.staticfile.on": True,
            "tools.staticfile.filename": "/tmp/ot2-favicon.png",
        },
    }
)


class KillerThread(Thread):
    def __init__(self, delay: float = 1):
        super(KillerThread, self).__init__()
        self._t = delay
    
    def run(self):
        time.sleep(self._t)
        os.kill(os.getpid(), 9)  # 9 -> SIGKILL


class StationRESTServer:
    def __init__(self, ctx: ProtocolContext, station: Optional['Station'] = None, config: Optional[dict] = None, favicon_url: Optional[str] = None):
        super(StationRESTServer, self).__init__()
        self._ctx = ctx
        self._station = station
        self._config = config
        self._icon_url = favicon_url
        self._status = None
    
    @cherrypy.expose
    def log(self) -> str:
        lws_logger = getattr(self._station, "_lws_logger", None)
        if lws_logger:
            ip = cherrypy.request.remote.ip
            try:
                ip = ipaddress.ip_address(ip)
            except ValueError:
                pass
            else:
                ipv4 = ip.ipv4_mapped if isinstance(ip, ipaddress.IPv6Address) else None
                ip = str(ip) if ipv4 is None else str(ipv4)
                if ip == "::1":
                    ip = "127.0.0.1"
            lws_logger.ip = ip
            self._station.logger.debug("Set runlog URL to: {}".format(lws_logger.url))
            
        if self._station._wait_first_log and self._station._waiting_first_log:
            self._station._ctx.resume()
            return json.dumps({})
        
        status = getattr(self._station, "status", None)
        tip_log = getattr(self._station, "_tip_log", {})

        tip_log_for_log = dict()
        for k in "count", "max":
            tip_log_for_log[k] = tip_log.get(k)

        # try:
        #     temp = getattr(getattr(self._station, "_tempdeck", None), "temperature", None)
        # except Exception:
        #     temp = None

        return json.dumps({
            "status": status if status == "finished" or self._status is None else self._status,
            "stage": getattr(self._station, "stage", None),
            "msg": getattr(self._station, "msg", None),
            "external": getattr(self._station, "external", False),
            "time": datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S:%f"),
            # "temp": temp,
            "tips": tip_log_for_log,
            "runlog": self._station._log_filepath,
        }, indent=2)
    
    @cherrypy.expose
    def pause(self):
        self._status = "pause"
        self._ctx.pause()
    
    @cherrypy.expose
    def resume(self):
        self._status = None
        self._ctx.resume()
    
    @cherrypy.expose
    def kill(self, delay: str = '1'):
        KillerThread(delay=float(delay)).start()

    @staticmethod
    def stop():
        cherrypy.engine.exit()


class StationRESTServerThread(StationRESTServer, Thread):
    def run(self):
        if self._icon_url and self._config and "tools.staticfile.filename" in self._config.get("/favicon.ico", {}):
            os.system("wget {} -O {}".format(self._icon_url, self._config["/favicon.ico"]["tools.staticfile.filename"]))
        cherrypy.quickstart(self, config=self._config)
    
    def join(self, timeout=None, after: float = 0):
        if after:
            time.sleep(after)
        self.stop()
        super(StationRESTServerThread, self).join(timeout=timeout)


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



from . import __version__, __file__ as module_path
from .request import StationRESTServerThread, DEFAULT_REST_KWARGS
from .utils import ProtocolContextLoggingHandler, LocalWebServerLogger
from .lights import Button, BlinkingLightHTTP, BlinkingLight
from opentrons.protocol_api import ProtocolContext
from opentrons.protocol_api.paired_instrument_context import PairedInstrumentContext
from opentrons.types import Point
import opentrons.commands.types
from abc import ABCMeta, abstractmethod
from functools import wraps, partialmethod
from itertools import chain, dropwhile
from opentrons.types import Location
from typing import Optional, Callable, Tuple
import json
import math
import os
import logging
import time


def loader(key):
    def loader_(idx: int = 0, *items: tuple) -> Callable:
        def _labware_loader(method: Callable) -> Callable:
            @wraps(method)
            def method_(self, *args, **kwargs):
                if items:
                    self.logger.debug("loading {}".format(", ".join(map(str, items))))
                return method(self, *args, **kwargs)
            
            setattr(method_, key, (idx, items))
            return method_
        return _labware_loader
    return loader_


# Decorator for labware loader methods
labware_loader = loader("_labware_load")
# Decorator for instrument loader methods
instrument_loader = loader("_instr_load")


def wells(rack):
    return rack.wells()


def first_row(rack):
    return rack.rows()[0]


class StationMeta(ABCMeta):
    def __new__(meta, name, bases, classdict):
        c = super(StationMeta, meta).__new__(meta, name, bases, classdict)
        try:
            with open(os.path.join(os.path.dirname(module_path), 'msg', '{}.json'.format(name))) as f:
                c._messages = json.load(f)
        except FileNotFoundError:
            c._messages = {}
        return c
    
    def get_message(cls, key: str, lan: str = 'ENG'):
        for c in cls.__mro__:
            d = getattr(c, '_messages', {})
            if key in d:
                d = d[key]
                if lan in d:
                    return d[lan]
        return key


class Station(metaclass=StationMeta):
    _protocol_description = "[BRIEFLY DESCRIBE YOUR PROTOCOL]"
    
    def __init__(self,
        debug_mode = False,
        drop_height: float = -2,
        drop_loc_l: float = 0,
        drop_loc_r: float = 0,
        drop_loc_y: float = 0,
        drop_threshold: int = 296,
        dummy_lights: bool = True,
        jupyter: bool = True,
        log_filepath: Optional[str] = '/var/lib/jupyter/notebooks/outputs/run_{}.log',
        log_file_format: Optional[str] = '%(asctime)s %(levelname)s %(message)s',
        log_file_date_format: Optional[str] = '%Y-%m-%d %H:%M:%S',
        log_lws_enable: Optional[str] = True,
        log_lws_ip: Optional[str] = None,
        log_lws_endpoint: str = ":5002/log",
        logger: Optional[logging.getLoggerClass()] = None,
        language: str = "ENG",
        metadata: Optional[dict] = None,
        num_samples: int = 96,
        rest_server_kwargs: dict = DEFAULT_REST_KWARGS,
        samples_per_col: int = 8,
        skip_delay: bool = False,
        start_at: Optional[str] = None,
        simulation_log_file: bool = False,
        simulation_log_lws: bool = False,
        tip_log_filename: str = 'tip_log.json',
        tip_log_folder_path: str = '/var/lib/jupyter/notebooks/outputs',
        tip_track: bool = True,
        wait_first_log: bool = False,
        **kwargs,
    ):
        self._debug_mode = debug_mode
        self._drop_height = drop_height
        self._drop_loc_l = drop_loc_l
        self._drop_loc_r = drop_loc_r
        self._drop_loc_y = drop_loc_y
        self._drop_threshold = drop_threshold
        self._dummy_lights = dummy_lights
        self.jupyter = jupyter
        self._language = language
        self._log_filepath = log_filepath.format(time.strftime("%Y_%m_%d__%H_%M_%S"))
        self._log_file_format = log_file_format
        self._log_file_date_format = log_file_date_format
        self._log_lws_enable = log_lws_enable
        self._log_lws_ip = log_lws_ip
        self._log_lws_endpoint = log_lws_endpoint
        self._logger = logger
        self.metadata = metadata
        self._num_samples = num_samples
        self._rest_server_kwargs = rest_server_kwargs
        self._samples_per_col = samples_per_col
        self._start_at = start_at
        self._skip_delay = skip_delay
        self._tip_log_filename = tip_log_filename
        self._tip_log_folder_path = tip_log_folder_path
        self._tip_track = tip_track
        self._ctx: Optional[ProtocolContext] = None
        self._drop_count = 0
        self._side_switch = True
        self._simulation_log_file = simulation_log_file
        self._simulation_log_lws = simulation_log_lws
        self._wait_first_log = wait_first_log
        self._waiting_first_log = False
        self.status = "initializing"
        self.stage = None
        self._msg = ""
        self.external = False
        self._run_stage = self._start_at is None
    
    def set_external(self, value: bool = True) -> bool:
        self.external = value
        return self.external
    
    set_internal = partialmethod(set_external, value=False)
    
    def get_msg(self, value: str) -> str:
        return str(type(self).get_message(value, self._language))
    
    def get_msg_format(self, value: str, *args, **kwargs) -> str:
        return self.get_msg(value).format(*args, **kwargs)
    
    @property
    def msg(self) -> str:
        return self._msg
    
    @msg.setter
    def msg(self, value: str):
        self._msg = self.get_msg(value)
    
    def msg_format(self, value: str, *args, **kwargs) -> str:
        self._msg = self.get_msg_format(value, *args, **kwargs)
        return self.msg
    
    def run_stage(self, stage: str) -> bool:
        self.stage = stage
        if self._start_at == self.stage:
            self._run_stage = True
        self.logger.info("[{}] Stage: {}".format("x" if self._run_stage else " ", self.stage))
        return self._run_stage
    
    @property
    def logger(self) -> logging.getLoggerClass():
        if ((not hasattr(self, "_logger")) or self._logger is None) and self._ctx is not None:
            self._logger = logging.getLogger(self.logger_name)
            if self._logger.hasHandlers():
                self._logger.handlers = []
            self._logger.addHandler(ProtocolContextLoggingHandler(self._ctx))
        return self._logger
    
    def setup_opentrons_logger(self):
        stack_logger = logging.getLogger('opentrons')
        stack_logger.setLevel(self.logger.getEffectiveLevel())
        if self._log_filepath and (self._simulation_log_file or not self._ctx.is_simulating()):
            os.makedirs(os.path.dirname(self._log_filepath), exist_ok=True)
            file_handler = logging.FileHandler(self._log_filepath)
            file_handler.setFormatter(logging.Formatter(self._log_file_format, self._log_file_date_format))
            stack_logger.addHandler(file_handler)

        if self._log_lws_enable:
            self._lws_logger = LocalWebServerLogger(self._log_lws_ip, self._log_lws_endpoint)
            if self._simulation_log_lws or not self._ctx.is_simulating():
                self._ctx.broker.subscribe(opentrons.commands.types.COMMAND, self._lws_logger)
    
    @property
    def logger_name(self) -> str:
        return self.__class__.__name__
    
    @classmethod
    def loaders(cls, key: str) -> map:
        return map(lambda x: (x[0], x[2]), sorted(map(lambda x: (x, *getattr(getattr(cls, x), key)), filter(lambda x: hasattr(getattr(cls, x), key), dir(cls))), key=lambda x: x[1]))
    
    @classmethod
    def labware_loaders(cls) -> map:
        return cls.loaders("_labware_load")
    
    @classmethod
    def instrument_loaders(cls) -> map:
        return cls.loaders("_instr_load")
    
    def load_it(self, it):
        for method_name, _ in it:
            getattr(self, method_name)()
    
    def load_labware(self):
        self.load_it(self.labware_loaders())
    
    def load_instruments(self):
        self.load_it(self.instrument_loaders())
        
    def equipment(self, it) -> dict:
        return {k: getattr(self, k) for _, v in it for k in v}
    
    @property
    def labware(self) -> dict:
        return self.equipment(self.labware_loaders())
    
    @property
    def instruments(self) -> dict:
        return self.equipment(self.instrument_loaders())
    
    @property
    def num_cols(self) -> int:
        return math.ceil(self._num_samples/self._samples_per_col)
    
    @property
    def _tip_log_filepath(self) -> str:
        return os.path.join(self._tip_log_folder_path, self._tip_log_filename)
    
    @abstractmethod
    def _tipracks(self) -> dict:
        """Mapping from tipracks attribute names to associated pipette (attribute names)"""
        pass

    def _update_tip_log_count(self):
        """Count the number of used tips"""
        self._tip_log['count'] = {t: sum(map(lambda x: not x.has_tip, self._tip_log['tips'][t])) for t in self._tipracks().keys()}

    def _reset_tips(self):
        for key, tips in self._tip_log['tips'].items():
            for t in tips:
                t.has_tip = True

    def setup_tip_log(self):
        data_tip_status = {}
        if self._tip_track:
            self.logger.info(self.msg_format("tip info log", self._tip_log_filepath))
            if os.path.isfile(self._tip_log_filepath):
                with open(self._tip_log_filepath) as json_file:
                    file = json.load(json_file)
                    data_tip_status: dict = file.get("tip_status", {})
        else:
            self.logger.debug("not using tip log file")
        
        self._tip_log = {
            'tips': {t: list(chain.from_iterable(map(first_row if getattr(self, p).channels > 1 else wells, getattr(self, t)))) for t, p in self._tipracks().items()}
        }
        if data_tip_status:
            for tiprack, tips in data_tip_status.items():
                assert len(tips) == len(self._tip_log['tips'][tiprack]), "Wrong length, please reset tip log"
                for i, tip in enumerate(tips):
                    assert tip.get("name", "") == str(self._tip_log['tips'][tiprack][i]), "Wrong tip order, please reset tip log"
                    self._tip_log['tips'][tiprack][i].has_tip = tip.get("has_tip", False)
        self._tip_log['max'] = {t: len(p) for t, p in self._tip_log['tips'].items()}
        self._update_tip_log_count()
    
    def track_tip(self):
        if self._tip_track and not self._ctx.is_simulating():
            self.logger.debug(self.get_msg_format("tip log dump", self._tip_log_filepath))
            os.makedirs(self._tip_log_folder_path, exist_ok=True)
            with open(self._tip_log_filepath, 'w') as outfile:
                json.dump({
                    "count": self._tip_log['count'],
                    "next": {k: str(self._get_next_tip(v)) for k, v in self._tip_log['tips'].items()},
                    "tip_status": {k: [{'name': str(w), 'has_tip': w.has_tip} for w in self._tip_log['tips'][k]] for k, v in self._tip_log['count'].items()}
                }, outfile, indent=2)

    @staticmethod
    def _get_next_tip(tips):
        available_tips = list(dropwhile(lambda x: not x.has_tip, tips))
        return available_tips[0] if len(available_tips) else None

    def pick_up(self, pip, loc: Optional[Location] = None, tiprack: Optional[str] = None):
        if loc is None:
            if tiprack is None:
                    for t in self._tipracks().keys():
                        if getattr(self, t) == pip.tip_racks:
                            tiprack = t
                            break
            if tiprack is None:
                if isinstance(pip, PairedInstrumentContext):
                    # Paired pipette tiprack doesn't work with the for loop below
                    # pipettes must be the same, so selecting the first tiprack
                    tiprack = list(self._tipracks().keys())[0]
                    self._logger.debug("Selectiong {} tiprack for paired pipette".format(tiprack))
                else:
                    raise RuntimeError("no tiprack associated to pipette {}".format(pip))
            self._update_tip_log_count()
            if self._tip_log['count'][tiprack] == self._tip_log['max'][tiprack]:
                # If empty, wait for refill
                self._reset_tips()
                self._update_tip_log_count()
                self.track_tip()
                self.pause(self.get_msg_format("refill tips", "\n".join(map(str, getattr(self, tiprack)))))
            self.track_tip()
            loc = self._get_next_tip(self._tip_log['tips'][tiprack])
        pip.pick_up_tip(loc)
        self._update_tip_log_count()
        self.track_tip()

    @staticmethod
    def _get_pipette_num_channels(pip):
        pips = pip._instruments.values() if isinstance(pip, PairedInstrumentContext) else [pip]
        return sum([p.channels for p in pips])

    def drop(self, pip):
        # Drop in the Fixed Trash (on 12) at different positions to avoid making a tall heap of tips
        drop_loc = self._ctx.loaded_labwares[12].wells()[0].top(self._drop_height).move(Point(x=self._drop_loc_r if self._side_switch else self._drop_loc_l, y=self._drop_loc_y))
        self._side_switch = not self._side_switch
        if self._debug_mode:
            pip.return_tip()
        else:
            pip.drop_tip(drop_loc)
        self._drop_count += self._get_pipette_num_channels(pip)
        if self._drop_count >= self._drop_threshold:
            self.pause('empty tips')
            self._drop_count = 0
    
    def pause(self,
        msg: str = "",
        blink: bool = True,
        blink_period: float = 1,
        color: str = 'red',
        delay_time: float = 0,
        home: bool = True,
        level: int = logging.INFO,
        pause: bool = True,
    ):
        self.status = "pause"
        old_color = self._button.color
        self._button.color = color
        if msg:
            self.msg = msg
            self.logger.log(level, self.msg)
        if home:
            self._ctx.home()
        if blink and not self._ctx.is_simulating():
            lt = (BlinkingLightHTTP if self._dummy_lights else BlinkingLight)(self._ctx, t=blink_period/2)
            lt.start()
        if delay_time > 0:
            self.status = "delay"
            self._ctx.delay(delay_time)
        if pause:
            self._ctx.pause()
            self._ctx.delay(0.1)  # pad to avoid pause leaking
        if blink and not self._ctx.is_simulating():
            lt.stop()
        self._button.color = old_color
        self.status = "running"
        self.msg = ""
    
    def dual_pause(self, msg: str, cols: Tuple[str, str] = ('red', 'yellow'), between: Optional[Callable] = None, home: Tuple[bool, bool] = (True, False)):
        msg = self.get_msg(msg)
        self._msg = "{}.\n{}".format(msg, self.get_msg("stop blink"))
        self.pause(self.msg, color=cols[0], home=home[0])
        if between is not None:
            between()
        self._msg = "{}.\n{}".format(msg, self.get_msg("continue"))
        self.pause(self.msg, blink=False, color=cols[1], home=home[1])
    
    def delay(self,
        mins: float,
        msg: str = "",
        color: str = 'yellow',
        home: bool = True,
        level: int = logging.INFO,
    ):
        self.pause(
            msg=self.get_msg_format("delay minutes", self.get_msg(msg), mins, self.get_msg("skip delay") if self._skip_delay else ""),
            blink=False,
            color=color,
            delay_time=0 if self._skip_delay else (60 * mins),
            home=home,
            level=level,
            pause=self._skip_delay,
        )
        
    def body(self):
        pass
    
    def run(self, ctx: ProtocolContext):
        self.status = "running"
        self._ctx = ctx
        self._button = (Button.dummy if self._dummy_lights else Button)(self._ctx, 'blue')
        if self._simulation_log_lws or not self._ctx.is_simulating():
            self._request = StationRESTServerThread(ctx, station=self, **self._rest_server_kwargs)
            self._request.start()
        
        self.setup_opentrons_logger()
        if self._wait_first_log:
            self._waiting_first_log = True
            self.pause("wait log", blink=False, home=False, color='yellow')
            self._waiting_first_log = False
        
        self.logger.info(self.msg_format("protocol description"))
        self.logger.info(self.msg_format("num samples", self._num_samples))
        self.logger.info(self.msg_format("version", __version__))
        
        self.load_labware()
        self.load_instruments()
        self.setup_tip_log()
        self._button.color = 'white'
        self.msg = ""

        if self._debug_mode:
            self.dual_pause("debug mode")
        
        try:
            self.body()
        finally:
            self.status = "finished"
            if not self._ctx.is_simulating():
                self._request.join(2, 0.5)
            self.track_tip()
            self._button.color = 'blue'
        self._ctx.home()
    
    def simulate(self):
        from opentrons import simulate
        self.run(simulate.get_protocol_api(self.metadata["apiLevel"]))


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



from ..multi_tube_source import MultiTubeSource
from ..utils import uniform_divide
from .technogenetics import StationCTechnogenetics
import logging


class StationCBioerMastermixPrep(StationCTechnogenetics):
    _protocol_description = "Bioer mastermix preparation protocol"

    def __init__(self,
            aspirate_rate: float = 30,
            debug_mode: bool = False,
            tube_bottom_headroom_height: float = 4.0,
            strip_bottom_headroom_height: float = 4.0,
            pcr_bottom_headroom_height: float = 4.5,
            dispense_rate: float = 30,
            mastermix_vol: float = 20,
            mastermix_vol_headroom: float = 60,
            mastermix_headroom_part_in_strip: float = 0.7,
            mm_tube_capacity = 1800,
            num_samples: int = 96,
            control_well_positions = ['A12', 'H12'],
            skip_delay: bool = False,
            source_plate_name: str = 'Clean PCR plate on block ',
            tube_block_model: str = "opentrons_24_aluminumblock_nest_1.5ml_screwcap",
            ** kwargs

        ):
        """ Build a :py:class:`.StationC`.
        :param aspirate_rate: Aspiration rate in uL/s
        :param tube_bottom_headroom_height: Height to keep from the bottom for mastermix tubes
        :param strip_bottom_headroom_height: Height to keep from the bottom for the strips
        :param pcr_bottom_headroom_height: Height to keep from the bottom for the output pcr plate
        :param dispense_rate: Dispensation rate in uL/s
        :param drop_loc_l: offset for dropping to the left side (should be positive) in mm
        :param drop_loc_r: offset for dropping to the right side (should be negative) in mm
        :param drop_threshold: the amount of dropped tips after which the run is paused for emptying the trash
        :param logger: logger object. If not specified, the default logger is used that logs through the ProtocolContext comment method
        :param mastermix_vol: Mastermix volume per sample in uL
        :param mastermix_vol_headroom: Headroom for mastermix preparation volume to add to needed volume
        :param metadata: protocol metadata
        :param num_samples: The number of samples that will be loaded on the station B
        :param control_well_positions: Position of the control wells to be filled with mastermix
        :param samples_per_col: The number of samples in a column of the destination plate
        :param source_plate_name: Name for the source plate
        :param skip_delay: If True, pause instead of delay.
        :param tube_block_model: Tube block model name
        """
        super(StationCBioerMastermixPrep, self).__init__(
            mm_tube_capacity = mm_tube_capacity,
            num_samples = num_samples,
            skip_delay = skip_delay,
            tube_block_model = tube_block_model,
            source_plate_name = source_plate_name,
            aspirate_rate = aspirate_rate,
            dispense_rate = dispense_rate,
            ** kwargs
        )
        self._debug_mode = debug_mode
        self._tube_bottom_headroom_height = tube_bottom_headroom_height
        self._strip_bottom_headroom_height = strip_bottom_headroom_height
        self._pcr_bottom_headroom_height = pcr_bottom_headroom_height
        self._mastermix_vol = mastermix_vol
        self._mastermix_vol_headroom = mastermix_vol_headroom

        assert 0 <= mastermix_headroom_part_in_strip <= 1, \
            "Mastermix headroom in strip part must be between or equal to 0 and 1"
        self._mastermix_headroom_part_in_strip = mastermix_headroom_part_in_strip
        self._control_well_positions = control_well_positions

        self._remaining_samples = self._num_samples
        self._done_cols: int = 0
        self._mastermix_volume_in_tip = 0       # Volume of mastermix passed present in p300 tip before distributing to strips
        self._mm_sources = MultiTubeSource("Mastermix tubes", self.logger)

    @property
    def sample_dests_wells(self):
        return self._pcr_plate.wells()[:self.num_cols*8]

    @property
    def remaining_cols(self):
        return self.num_cols - self._done_cols

    @property
    def mm_strip(self):
        # We use only one column
        return self._mm_strips.columns()[0]

    @property
    def headroom_from_strip_to_pcr(self):
        return ((self._mastermix_vol_headroom - 1.0) / 2) + 1.0

    @property
    def headroom_vol_from_strip_to_pcr(self):
        return self._mastermix_vol_headroom * self._mastermix_headroom_part_in_strip

    @property
    def headroom_vol_from_strip_to_pcr_single(self):
        return self.headroom_vol_from_strip_to_pcr / 8

    @property
    def headroom_vol_from_tubes_to_strip(self):
        return self._mastermix_vol_headroom * (1 - self._mastermix_headroom_part_in_strip)

    @property
    def control_dests_wells(self):
        return [self._pcr_plate.wells_by_name()[i] for i in self._control_well_positions]  # controlli in posizione A12 e H12

    def is_well_in_samples(self, well):
        """
        Function that check if a well is within the samples well.
        :param well: well to check
        :return: True if the well is included in the samples list.
        """
        return well in self.sample_dests_wells

    @property
    def control_wells_not_in_samples(self):
        """
        :return: a list of wells for controls that are not already filled with the 8-channel pipette
        """
        return [c for c in self.control_dests_wells if not self.is_well_in_samples(c)]

    def fill_strip(self, volume):
        if not self._p300.has_tip:
            self.pick_up(self._p300)

        total_vol = volume * 8
        self.logger.debug("Filling strips with {}ul each; used volume: {}".format(volume, total_vol))

        for well in self.mm_strip:
            self._mm_sources.aspirate_from_tubes(volume - self._mastermix_volume_in_tip, self._p300, self._tube_bottom_headroom_height)
            self._mastermix_volume_in_tip = 0   # only doing the first time
            self._p300.dispense(volume, well.bottom(self._strip_bottom_headroom_height))

        self.drop(self._p300)


    def fill_controls(self, dead_volume: float = 0):

        if len(self.control_wells_not_in_samples) > 0:
            self.logger.info(self.msg_format("fill control", self.control_wells_not_in_samples))
            if not self._p300.has_tip:
                self.pick_up(self._p300)

            vol = self._mastermix_vol * len(self.control_wells_not_in_samples) + dead_volume
            self._mm_sources.aspirate_from_tubes(vol, self._p300, self._tube_bottom_headroom_height)

            for w in self.control_dests_wells:
                self._p300.dispense(self._mastermix_vol, w.bottom(self._pcr_bottom_headroom_height))

            assert self._p300.current_volume == dead_volume, \
                "Please verify: dead volume in tip is not as expected. in tip {}, expected {}".format(
                    self._p300.current_volume, dead_volume)

            self._mastermix_volume_in_tip = dead_volume     # Saving volume, eventually it will be disposed to strip...
        else:
            self.logger.info(self.get_msg("controls already filled"))

    def transfer_to_pcr_plate_and_mark_done(self, num_columns: int):
        num_columns = int(num_columns)
        self.logger.info("Transferring to pcr place {:d} columns.".format(num_columns))
        self.pick_up(self._m20)
        for s in self.get_next_pcr_plate_dests(num_columns):
            self._m20.transfer(self._mastermix_vol,
                               self.mm_strip[0].bottom(self._strip_bottom_headroom_height),
                               s.bottom(self._pcr_bottom_headroom_height),
                               new_tip='never')
        self.drop(self._m20)

    def get_next_pcr_plate_dests(self, num_columns: int):
        if self._done_cols < self.num_cols:
            to_do = int(min(self.num_cols - self._done_cols, num_columns))
        else:
            raise Exception("No more columns to do")
        samples_to_do = self.sample_dests[self._done_cols:self._done_cols+to_do]
        self._done_cols += to_do
        return samples_to_do

    def body(self):
        volume_for_controls = len(self.control_wells_not_in_samples) * self._mastermix_vol
        volume_for_samples = self._mastermix_vol * self.num_cols * 8
        volume_to_distribute_to_pcr_plate = volume_for_samples + volume_for_controls
        volume_to_distribute_to_strip = volume_for_samples + self.headroom_vol_from_strip_to_pcr

        total_volume = volume_to_distribute_to_strip + volume_for_controls + self.headroom_vol_from_tubes_to_strip
        self.logger.info("For this run we need a total of {}ul of mastermix".format(total_volume))

        num_tubes, vol_per_tube = uniform_divide(total_volume, self._mm_tube_capacity)
        mm_tubes = self._tube_block.wells()[:num_tubes]

        # Filling source class to calculate where to aspirate
        available_volume = (volume_to_distribute_to_strip + volume_for_controls) / len(mm_tubes)
        assert vol_per_tube > available_volume, \
            "Error in volume calcuations: requested {}ul while total in tubes {}ul".format(available_volume,
                                                                                           vol_per_tube)
        for source in mm_tubes:
            self._mm_sources.append_tube_with_vol(source, available_volume)

        self.logger.info("")
        self.logger.info("We need {} tubes with {}ul of mastermix each in {}.".format(num_tubes,
                                                                                      vol_per_tube,
                                                                                      self._mm_sources))
        self.logger.debug("{}ul will be dispensed to control positions.".format(volume_for_controls))
        self.logger.debug("{}ul will be dispensed to PCR plate".format(volume_to_distribute_to_pcr_plate))
        self.logger.debug("{}ul will be dispensed to strips".format(volume_to_distribute_to_strip))
        self.logger.debug("")
        self.logger.debug("In this run we use a volume overhead of: {}ul".format(self._mastermix_vol_headroom))
        self.logger.debug("\t- {:.1f}ul will remain in tubes".format(self.headroom_vol_from_tubes_to_strip))
        self.logger.debug("\t- {:.1f}ul will remain in strips".format(self.headroom_vol_from_strip_to_pcr))

        # Variable to transfer headroom to strip only the first time
        strip_headroom_vol_single_first_time = self.headroom_vol_from_strip_to_pcr_single

        # First fill controls
        self.fill_controls(dead_volume=self._p300.min_volume)

        # Main loop filling the plate
        while self.remaining_cols > 0:
            # calcuations for filling strip each time
            self.logger.debug("Remaining columns: {}".format(self.remaining_cols))
            strip_volume = min(self._mm_strip_capacity,
                               self.remaining_cols * self._mastermix_vol + strip_headroom_vol_single_first_time)

            samples_per_this_strip = (strip_volume - strip_headroom_vol_single_first_time) // self._mastermix_vol


            strip_fill_volume = samples_per_this_strip * self._mastermix_vol + strip_headroom_vol_single_first_time
            strip_headroom_vol_single_first_time = 0 # resetting the headroom volume, will still be present in strips

            self.logger.debug("Filling strip with {}ul and using it for {} samples".format(strip_fill_volume,
                                                                                          samples_per_this_strip))
            self.fill_strip(strip_fill_volume)

            self.transfer_to_pcr_plate_and_mark_done(samples_per_this_strip)

        if self._p300.has_tip:
            self.drop(self._p300)

        self.logger.debug("Remaining vols: {}".format(self._mm_sources))

    def drop(self, pip):
        if self._debug_mode:
            pip.return_tip()
        else:
            super(StationCBioerMastermixPrep, self).drop(pip)

arguments = dict(num_samples=88, debug_mode=False)

# protocol for loading in Opentrons App or opentrons_simulate
# =====================================
logging.getLogger(StationCBioerMastermixPrep.__name__).setLevel(logging.INFO)
metadata = {'apiLevel': '2.7'}
station = StationCBioerMastermixPrep(**arguments)


def run(ctx):
    return station.run(ctx)


# for running directly with python command 'py Mastermix_prep_stations.py"
# ========================================================================
if __name__ == "__main__":
    StationCBioerMastermixPrep(**arguments, metadata={'apiLevel': '2.7'}).simulate()



from .p1000 import StationAP1000
from .reload import StationAReloadMixin
from .copan_24 import Copan24Specs
from .copan_48 import copan_48_corrected_specs
from typing import Tuple, Optional


class StationATechnogenetics(StationAP1000):
    def __init__(self,
        beads_mix_repeats: int = 2,
        beads_mix_volume: float = 20,
        beads_vol: float = 10,
        drop_threshold: int = 95,
        lysis_first: bool = False,
        lys_mix_repeats: int = 2,
        lys_mix_volume: float = 100,
        lysis_volume: float = 400,
        mix_repeats: int = 1,
        prot_k_capacity: float = 180,
        prot_k_headroom: float = 1.1,
        prot_k_vol: float = 30,
        sample_aspirate: float = 100,
        sample_dispense: float = 100,
        tempdeck_temp: Optional[float] = None,
        tempdeck_bool: bool = False,
        tipracks_slots: Tuple[str, ...] = ('10', '11'),
        tipracks_slots_20: Tuple[str, ...] = ('8', '9'),
        touch_tip_height: float = -5,
        *args,
        **kwargs
    ):
        """
        :param beads_mix_repeats: number of repetitions during mixing the beads
        :param beads_mix_volume: volume aspirated for mixing the beads in uL
        :param beads_vol: volume of beads per sample in uL
        :param lysis_volume: volume of lysis buffer per sample in uL
        :param prot_k_headroom: headroom for proteinase K (as a multiplier)
        :param prot_k_vol: volume of proteinase K per sample in uL
        :param sample_aspirate: aspiration rate for sampeles in uL/s
        :param sample_dispense: dispensation rate for sampeles in uL/s
        :param tempdeck_temp: tempdeck temperature in Celsius degrees
        :param tipracks_slots: Slots where the tipracks are positioned
        :param tipracks_slots_20: Slots where the tipracks (20 uL) are positioned
        :param kwargs: other keyword arguments. See: StationAP1000, StationA
        """
        super(StationATechnogenetics, self).__init__(
            *args,
            lysis_first=False,
            lys_mix_repeats=lys_mix_repeats,
            lys_mix_volume=lys_mix_volume,
            lysis_volume=lysis_volume,
            mix_repeats=mix_repeats,
            sample_aspirate=sample_aspirate,
            sample_dispense=sample_dispense,
            iec_volume=prot_k_vol,
            ic_capacity=prot_k_capacity,
            ic_lys_headroom=prot_k_headroom,
            tempdeck_temp=tempdeck_temp,
            tipracks_slots=tipracks_slots,
            tipracks_slots_20=tipracks_slots_20,
            **kwargs
        )
        self._beads_mix_repeats = beads_mix_repeats
        self._beads_mix_volume = beads_mix_volume
        self._beads_vol = beads_vol
        self._drop_threshold = drop_threshold
        self._tempdeck_bool = tempdeck_bool
        self._touch_tip_height = touch_tip_height
        if self._lysis_first != lysis_first:
            self.logger.error("lysis_first=True is not supported for this protocol")
    
    # --- In Station A there is IC in the strips, these allow the renaming for Proteinase K -------
    _strips_content: str = "proteinase K"
    
    _lys_buf_name: str = '50ml tuberack for lysis buffer'
    
    @property
    def _prot_k_capacity(self) -> float:
        return self._ic_capacity
    
    @property
    def _prot_k_headroom(self) -> float:
        return self._ic_lys_headroom
    
    @property
    def _prot_k_volume(self) -> float:
        return self._iec_volume
            
    @property
    def num_pk_strips(self) -> int:
        return self.num_ic_strips
    # ---------------------------------------------------------------------------------------------
    
    @property
    def _prot_k(self):
        return self._strips_block.rows()[0][:self.num_pk_strips]
    
    @property
    def _beads(self):
        return self._strips_block.rows()[0][-1]
    
    def transfer_proteinase(self):
        for i, d in enumerate(self._dests_multi):
            if self.run_stage("transfer proteinase {}/{}".format(i + 1, len(self._dests_multi))):
                self.pick_up(self._m20)
                self._m20.transfer(self._prot_k_volume, self._prot_k[i // self.cols_per_strip], d.bottom(self._ic_headroom_bottom), new_tip='never', touch_tip=True)
                #self._m20.touch_tip(v_offset=self._touch_tip_height)
                self._m20.drop_tip()
    
    def transfer_beads(self):
        for i, d in enumerate(self._dests_multi):
            if self.run_stage("transfer beads {}/{}".format(i + 1, len(self._dests_multi))):
                self.pick_up(self._m20)
                self._m20.transfer(self._beads_vol, self._beads, d.bottom(self._dest_multi_headroom_height), air_gap=self._air_gap_dest_multi, new_tip='never')
                self._m20.mix(self._beads_mix_repeats, self._beads_mix_volume, d.bottom(self._dest_multi_headroom_height))
                self._m20.air_gap(self._air_gap_dest_multi)
                self._m20.drop_tip()
    
    def body(self):
        self.setup_samples()
        self.setup_lys_tube()
        self.msg = ""
        
        self.transfer_proteinase()
        if self.run_stage("check proteinase"):
            self.dual_pause("proteinase check")
        self.transfer_samples()
        self.transfer_lys()
        
        if self.run_stage("incubation"):
            self.dual_pause("incubate", between=self.set_external)
            self.set_internal()
        
        self.transfer_beads()
        self.logger.info(self.msg_format("move to B"))


class StationATechnogeneticsReload(StationAReloadMixin, StationATechnogenetics):
    _protocol_description = "station A protocol for Technogenetics kit and COPAN 330C refillable samples."


class StationATechnogenetics24(StationATechnogenetics):
    _protocol_description = "station A protocol for Technogenetics kit and COPAN 330C (x24 rack)"

    def _load_source_racks(self):
        labware_def = Copan24Specs().labware_definition()
        self._source_racks = [
            self._ctx.load_labware_from_definition(
                labware_def, slot,
                'source tuberack ' + str(i + 1)
            ) for i, slot in enumerate(self._source_racks_slots)
        ]


class StationATechnogenetics48(StationATechnogeneticsReload):
    _protocol_description = "station A protocol for Technogenetics kit and COPAN 330C (x48 rack)"
    
    def __init__(
        self,
        positive_control_well: str = 'H12',
        source_racks_slots: Tuple[str, ...] = ('2',),
        *args,
        **kwargs
    ):
        super(StationATechnogenetics48, self).__init__(
            *args,
            positive_control_well=positive_control_well,
            source_racks_slots=source_racks_slots,
            **kwargs
        )
    
    def _load_source_racks(self):
        labware_def = copan_48_corrected_specs.labware_definition()
        self._source_racks = [
            self._ctx.load_labware_from_definition(
                labware_def, slot,
                'source tuberack ' + str(i + 1)
            ) for i, slot in enumerate(self._source_racks_slots)
        ]


if __name__ == "__main__":
    StationATechnogenetics24(num_samples=96, metadata={'apiLevel': '2.3'}).simulate()


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



import json
from covmatic_stations.station import instrument_loader, labware_loader
from covmatic_stations.utils import uniform_divide
from .technogenetics import StationBTechnogenetics
from ..paired_pipette import PairedPipette
from opentrons.types import Point
from itertools import repeat
import math
import os

class StationBTechnogeneticsPairedPipette(StationBTechnogenetics):

    def __init__(self, **kwargs):
        super(StationBTechnogeneticsPairedPipette, self).__init__(**kwargs)

    @labware_loader(8, "_res12")
    def load_res12(self):
        # re-loading labware to avoid bug on reagent reservoir (see https://github.com/Opentrons/opentrons/issues/7793)
        reservoir_file = os.path.join(os.path.split(__file__)[0], "nest_12_reservoir_15ml_modified.json")
        self.logger.info("Loading reservoir labware from: {}".format(reservoir_file))
        with open(reservoir_file) as labware_file:
            labware_def = json.load(labware_file)
        self._res12 = self._ctx.load_labware_from_definition(labware_def, 5, 'Trough with WashReagents')

    @property
    def wash1(self):
        return self._res12.rows()[0][:6]

    @property
    def wash2(self):
        return self._res12.rows()[0][-6:]

    @property
    def water(self):
        return self._elut12.rows()[0][11]

    @instrument_loader(0, "_m300r")
    def load_m300_right(self):
        self._m300r = self._ctx.load_instrument('p300_multi_gen2', 'right', tip_racks=self._tips300)
        self.logger.info("Loaded right pipette")
        if self._bind_aspiration_rate:
            self._m300r.flow_rate.aspirate = self._bind_aspiration_rate
        if self._bind_dispense_rate:
            self._m300r.flow_rate.dispense = self._bind_dispense_rate
        if self._bind_blowout_rate:
            self._m300r.flow_rate.blow_out = self._bind_blowout_rate

    def body(self):
        PairedPipette.setup(self._m300, self._m300r, self)
        super(StationBTechnogeneticsPairedPipette, self).body()

    def mix_samples(self):
        self._m300.flow_rate.aspirate = 94
        self._m300r.flow_rate.aspirate = 94

        with PairedPipette(self._magplate, self.mag_samples_m, start_at="mix sample") as tp:
            tp.pick_up()
            tp.mix(self._sample_mix_times, self._sample_mix_vol,
                   locationFrom="target",
                   well_modifier="bottom({})".format(self._sample_mix_height))
            tp.air_gap(self._bind_air_gap)
            tp.drop_tip()


    def remove_supernatant(self, vol: float, stage: str = "remove supernatant"):
        self._m300.flow_rate.aspirate = self._supernatant_removal_aspiration_rate
        self._m300r.flow_rate.aspirate = self._supernatant_removal_aspiration_rate

        num_trans = math.ceil((vol - self._vol_last_trans) / self._bind_max_transfer_vol)
        vol_per_trans = ((vol - self._vol_last_trans) / num_trans) if num_trans else 0

        waste_locs = list(repeat(self._waste, len(self.mag_samples_m)))

        with PairedPipette(self._magplate, self.mag_samples_m, waste_locs=waste_locs, start_at=stage) as tp:
            tp.pick_up()
            for i in range(num_trans):
                # tp.move_to()  # we want center() but for now it is not in available commands
                if i > 0:
                    tp.dispense(self._supernatant_removal_air_gap, locationFrom="target", well_modifier="top(0)")
                tp.aspirate(vol_per_trans, locationFrom="target", well_modifier="bottom({})".format(self._supernatant_removal_height))
                tp.air_gap(self._supernatant_removal_air_gap)
                tp.dispense(vol_per_trans+self._supernatant_removal_air_gap, locationFrom="waste_locs")
                tp.air_gap(self._supernatant_removal_air_gap)

            tp.comment("Supernatant removal: last transfer in {} step".format(self._n_bottom))
            tp.dispense(self._supernatant_removal_air_gap, locationFrom="target", well_modifier="top(0)")
            for i in range(self._n_bottom):
                aspirate_height = self._h_bottom - i * (self._h_bottom / (self._n_bottom - 1))
                tp.comment("Aspirating at {}".format(aspirate_height))
                tp.aspirate(self._supernatant_removal_last_transfer_max_vol/self._n_bottom,
                            locationFrom="target",
                            well_modifier="bottom({})".format(aspirate_height))

            back_step = 0.1
            n_back_step = 3
            for _ in range(n_back_step):
                aspirate_height = aspirate_height + back_step
                tp.comment("Moving up at {}".format(aspirate_height))
                tp.move_to(locationFrom="target", well_modifier="bottom({})".format(aspirate_height))

            tp.air_gap(self._supernatant_removal_air_gap)
            tp.dispense(self._supernatant_removal_last_transfer_max_vol + self._supernatant_removal_air_gap,
                        locationFrom="waste_locs")
            tp.air_gap(self._supernatant_removal_air_gap)
            tp.drop_tip()

        self._m300.flow_rate.aspirate = self._default_aspiration_rate
        self._m300r.flow_rate.aspirate = self._default_aspiration_rate

    def elute(self, positions=None, transfer: bool = False, stage: str = "elute"):
        if positions is None:
            positions = self.temp_samples_m
        self._magdeck.disengage()
        self._elute(positions=positions, transfer=transfer, stage=stage)
        self._magdeck.disengage()

    def _elute(self, positions=None, transfer: bool = True, stage: str = "elute"):
        """Resuspend beads in elution"""
        # if positions is None:
        #     positions = self.mag_samples_m
        self._m300.flow_rate.aspirate = self._elute_aspiration_rate

        side_locs = []
        for i, m in enumerate(positions):
            side = 1 if i % 2 == 0 else -1
            side_locs.append(m.bottom(self._bottom_headroom_height).move(Point(x=side * 2)))

        with PairedPipette(self._tempplate, positions, side_locs=side_locs, start_at=stage) as tp:
            tp.pick_up()
            tp.aspirate(self._elution_vol, self.water)
            tp.air_gap(self._elute_air_gap)
            tp.dispense(self._elute_air_gap, locationFrom="target", well_modifier="top(0)")
            tp.dispense(self._elution_vol, locationFrom="side_locs")
            tp.mix(self._elute_mix_times, self._elute_mix_vol, locationFrom="side_locs")
            tp.touch_tip(v_offset=self._touch_tip_height)
            tp.air_gap(self._elute_air_gap)
            tp.drop_tip()

    def final_transfer(self):
        self._m300.flow_rate.aspirate = self._final_transfer_rate_aspirate
        self._m300r.flow_rate.aspirate = self._final_transfer_rate_aspirate
        self._m300.flow_rate.dispense = self._final_transfer_rate_dispense
        self._m300r.flow_rate.dispense = self._final_transfer_rate_dispense

        n = len(list(zip(self.mag_samples_m, self.pcr_samples_m)))

        locs = []
        for i, (m, e) in enumerate(zip(self.mag_samples_m, self.pcr_samples_m)):
            side = -1 if i % 2 == 0 else 1
            locs.append(m.bottom(0.3).move(Point(x=side * 2)))

        with PairedPipette(self._magplate, locs, start_at="final transfer", pcr_locs=self.pcr_samples_m) as tp:
            tp.pick_up()
            tp.aspirate(self._final_vol, locationFrom="target")
            tp.air_gap(self._elute_air_gap)
            tp.dispense(self._elute_air_gap, locationFrom="pcr_locs", well_modifier="top(0)")
            tp.dispense(self._final_vol, locationFrom="pcr_locs", well_modifier="bottom(0.5)")
            tp.mix(self._final_mix_times, self._final_mix_vol, locationFrom="pcr_locs",
                   well_modifier="bottom({})".format(self._final_mix_height))
            tp.air_gap(self._elute_air_gap)
            tp.drop_tip()

    def wash(self, vol: float, source, mix_reps: int, wash_name: str = "wash"):
        self.logger.info(self.msg_format("wash info", vol, wash_name, mix_reps))

        if wash_name == "wash 1":
            dispense_rate = self._wash_1_mix_dispense_rate
            self._m300.flow_rate.dispense = self._wash_1_mix_dispense_rate
            self._m300r.flow_rate.dispense = self._wash_1_mix_dispense_rate
            self._m300.flow_rate.aspirate = self._wash_1_mix_aspiration_rate
            self._m300r.flow_rate.aspirate = self._wash_1_mix_aspiration_rate
        else:
            dispense_rate = self._wash_2_mix_dispense_rate
            self._m300.flow_rate.dispense = self._wash_2_mix_dispense_rate
            self._m300r.flow_rate.dispense = self._wash_2_mix_dispense_rate
            self._m300.flow_rate.aspirate = self._wash_2_mix_aspiration_rate
            self._m300r.flow_rate.aspirate = self._wash_2_mix_aspiration_rate
        self._magdeck.disengage()

        num_trans, vol_per_trans = uniform_divide(vol, self._wash_max_transfer_vol)

        src = []
        mix_locs = []
        wash_locs = []
        wash_locs_index = [0, 0, 1, 1, 4, 4, 5, 5, 2, 2, 3, 3]

        for i, m in enumerate(self.mag_samples_m):
            wash_locs.append(source[wash_locs_index[i]])
            src.append(self.wash_getcol(i, len(self.mag_samples_m), source))
            mix_locs.append(m.bottom(self._bottom_headroom_height).move(Point(x=2 * (-1 if i % 2 else +1))))

        with PairedPipette(self._magplate, self.mag_samples_m,
                           wash_locs=wash_locs,
                           mix_locs=mix_locs,
                           start_at=wash_name) as tp:
            tp.pick_up()
            for n in range(num_trans):
                if n > 0:
                    tp.dispense(self._wash_air_gap, locationFrom="wash_locs", well_modifier="top(0)")
                tp.aspirate(vol_per_trans, locationFrom="wash_locs")
                tp.air_gap(self._wash_air_gap)
                tp.dispense(vol_per_trans+self._wash_air_gap, locationFrom="target", well_modifier="top(0)")
                if n < num_trans-1:
                    tp.air_gap(self._wash_air_gap)

            # Mix
            tp.mix(mix_reps, self._wash_mix_vol, locationFrom="mix_locs")
            tp.air_gap(self._wash_air_gap)
            tp.drop_tip()

        self._magdeck.engage(height=self._magheight)
        self.check()
        if self.run_stage("{} incubate".format(wash_name)):
            self.delay(self._wait_time_wash_on, self.get_msg_format("incubate on magdeck", self.get_msg("on")))
        self.remove_supernatant(vol, stage="remove {}".format(wash_name))

if __name__ == "__main__":
    station = StationBTechnogeneticsPairedPipette(metadata={'apiLevel': '2.3'},
                                                  num_samples=96)
    # logging.getLogger(StationBTechnogeneticsPairedPipette.__name__).setLevel(logging.DEBUG)
    station.simulate()




from .technogenetics import StationBTechnogenetics
from typing import Tuple
from itertools import repeat


class StationBTechnogeneticsShort(StationBTechnogenetics):    
    def __init__(
        self,
        air_gap_drop: float = 0,
        bind_aspiration_rate=None,
        bind_dispense_rate=None,
        bind_blowout_rate=None,
        bottom_headroom_height: float = 0.1,
        elution_vol: float = 40,
        flatplate_slot: str = '3',
        magdeck_slot: str = '1',
        num_cycles: int = 1,
        tempdeck_temp = None,
        tipracks_slots: Tuple[str, ...] = ('4', '6', '7', '8', '9'),
        *args,
        **kwargs
    ):
        super(StationBTechnogeneticsShort, self).__init__(
            *args,
            bind_aspiration_rate=bind_aspiration_rate,
            bind_dispense_rate=bind_dispense_rate,
            bind_blowout_rate=bind_blowout_rate,
            bottom_headroom_height=bottom_headroom_height,
            elution_vol=elution_vol,
            flatplate_slot=flatplate_slot,
            magdeck_slot=magdeck_slot,
            tempdeck_temp=tempdeck_temp,
            tipracks_slots=tipracks_slots,
            **kwargs
        )
        self._air_gap_drop = air_gap_drop
        self._num_cycles = num_cycles

    def load_tempdeck(self):
        pass
    
    def load_tempplate(self):
        pass
    
    def load_res12(self):
        pass
    
    def load_elut12(self):
        pass
    
    @property
    def transfer_dest(self):
        return (e.bottom(self._elution_height) for e in self.pcr_samples_m)
    
    def cycle(self, idx: int, stage: str = "cycle"):
        if self.run_stage("{} {}/{}".format(stage, idx + 1, self._num_cycles)):
            self._magdeck.engage(height=self._magheight)
            self.delay(2, self.get_msg_format("incubate on magdeck", self.get_msg("on")))
            for i, (m, e) in enumerate(zip(self.mag_samples_m, self.transfer_dest)):
                self.pick_up(self._m300)
                self._m300.flow_rate.aspirate = self._elute_aspiration_rate
                self._m300.transfer(self._elution_vol, m.bottom(self._bottom_headroom_height), e, air_gap=self._elute_air_gap, new_tip='never')
                if self._air_gap_drop:
                    self._m300.air_gap(self._air_gap_drop)
                self.drop(self._m300)
            
            self._magdeck.disengage()
            
            if idx < self._num_cycles - 1:
                self.dual_pause("end of cycle")
            else:
                self.pause("final cycle")
        
    def body(self):
        for i in range(self._num_cycles):
            self.cycle(i)


class StationBTechnogeneticsElutionRemoval(StationBTechnogeneticsShort):
    _protocol_description = "elution removal protocol for station B with Technogenetics kit"
    
    def load_waste(self):
        pass


class StationBTechnogeneticsWashBRemoval(StationBTechnogeneticsShort):
    _protocol_description = "wash B removal protocol for station B with Technogenetics kit"
    
    def __init__(
        self,
        air_gap_drop: float = 20,
        wash_b_vol: float = 55,
        *args,
        **kwargs
    ):
        super(StationBTechnogeneticsWashBRemoval, self).__init__(
            *args,
            air_gap_drop=air_gap_drop,
            elution_vol=wash_b_vol,
            **kwargs
        )
    
    def load_flatplate(self):
        pass
    
    @property
    def transfer_dest(self):
        return repeat(self._waste)



import logging

class MultiTubeSource(object):
    """

    """
    def __init__(self, name="", logger=None):
        self._source_tubes_and_vol = []
        self._name = name
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(__name__)

    def append_tube_with_vol(self, source, available_volume):
        self._source_tubes_and_vol.append(dict(source=source,
                                               available_volume=available_volume))
        self.logger.debug("{}: appended {} with {}ul".format(self._name, source, available_volume))
        self.logger.debug("Now sources is: {}".format(self._source_tubes_and_vol))

    def aspirate_from_tubes(self, volume, pip, aspirate_height_from_bottom: float = 2):
        aspirate_list = []
        left_volume = volume
        for source_and_vol in self._source_tubes_and_vol:
            if source_and_vol["available_volume"] >= left_volume:
                # aspirate_list.append(dict(source=source_and_vol["source"], vol=left_volume))
                aspirate_vol = left_volume
            else:
                aspirate_vol = source_and_vol["available_volume"]
            left_volume -= aspirate_vol
            source_and_vol["available_volume"] -= aspirate_vol
            if aspirate_vol != 0:
                aspirate_list.append(dict(source=source_and_vol["source"], vol=aspirate_vol))

            if left_volume == 0:
                break
        else:
            raise Exception("{}: no volume left in source tubes.".format(self._name))

        for a in aspirate_list:
            pip.aspirate(a["vol"], a["source"].bottom(aspirate_height_from_bottom))

        self.logger.debug("{} sources: {}".format(self._name, self._source_tubes_and_vol))

    @property
    def locations_str(self):
        return "{}: ".format(self._name) + \
               " ".join(["{}; ".format(sv['source'])
                         for sv in self._source_tubes_and_vol])
    def __str__(self):
        return "{}: ".format(self._name) +\
               " ".join(["{} with volume {}ul;".format(sv['source'], sv['available_volume'])
                        for sv in self._source_tubes_and_vol])



from .b import StationB, labware_loader
from typing import Tuple
from ..utils import uniform_divide
from opentrons.types import Point


class StationBTechnogenetics(StationB):
    _protocol_description = "station B protocol for Technogenetics kit"

    def __init__(self,
                 elute_mix_times: int = 15,
                 elution_vol: float = 50,
                 elute_incubate: bool = False,
                 external_deepwell_incubation: bool = True,
                 final_mix_height: float = 0.3,
                 final_mix_times: int = 5,
                 final_mix_vol: float = 20,
                 final_transfer_rate_aspirate: float = 30,
                 final_transfer_rate_dispense: float = 30,
                 final_vol: float = 20,
                 flatplate_slot: str = '3',
                 h_bottom: float = 0.3,
                 n_bottom: float = 3,
                 mix_incubate_on_time: float = 20,
                 mix_incubate_off_time: float = 5,
                 postspin_incubation_time: float = 3,
                 remove_wash_vol: float = 50,
                 sample_mix_height: float = 0.3,
                 sample_mix_times: float = 10,
                 sample_mix_vol: float = 180,
                 starting_vol: float = 650,
                 tempdeck_slot: str = '10',
                 tempdeck_temp: float = 60,
                 thermomixer_incubation_time: float = 5,
                 tipracks_slots: Tuple[str, ...] = ('4', '6', '7', '8', '9'),
                 wash_1_vol: float = 680,
                 wash_2_vol: float = 680,
                 **kwargs
                 ):
        """ Build a :py:class:`.StationBTechnogenetics`.
        :param external_deepwell_incubation: whether or not to perform deepwell incubation outside the robot
        :param final_mix_height: Mixing height (from the bottom) for final transfer in mm
        :param final_mix_times: Mixing repetitions for final transfer
        :param final_mix_vol: Mixing volume for final transfer in uL
        :param final_transfer_rate_aspirate: Aspiration rate during final transfer in uL/s
        :param final_transfer_rate_dispense: Dispensation rate during final transfer in uL/s
        :param final_vol: Volume to transfer to the PCR plate in uL
        :param mix_incubate_on_time: Time for incubation on magnet after mix in minutes 
        :param mix_incubate_off_time: Time for incubation off magnet after mix in minutes
        :param postspin_incubation_time: Post-spin incubation time in minutes
        :param remove_wash_vol: Volume to remove during wash removal in uL
        :param sample_mix_times: Mixing height for samples in mm from the bottom
        :param sample_mix_times: Mixing repetitions for samples
        :param sample_mix_vol: Mixing volume for samples in uL
        :param thermomixer_incubation_time: Time for incubation after thermomixer in minutes 
        """
        super(StationBTechnogenetics, self).__init__(
            elute_mix_times=elute_mix_times,
            elution_vol=elution_vol,
            elute_incubate=elute_incubate,
            flatplate_slot=flatplate_slot,
            starting_vol=starting_vol,
            tempdeck_slot=tempdeck_slot,
            tempdeck_temp=tempdeck_temp,
            tipracks_slots=tipracks_slots,
            wash_1_vol=wash_1_vol,
            wash_2_vol=wash_2_vol,
            **kwargs
        )
        self._external_deepwell_incubation = external_deepwell_incubation
        self._final_mix_height = final_mix_height
        self._final_mix_times = final_mix_times
        self._final_mix_vol = final_mix_vol
        self._final_transfer_rate_aspirate = final_transfer_rate_aspirate
        self._final_transfer_rate_dispense = final_transfer_rate_dispense
        self._final_vol = final_vol
        self._flatplate_slot = flatplate_slot
        self._h_bottom = h_bottom
        self._n_bottom = n_bottom
        self._mix_incubate_on_time = mix_incubate_on_time
        self._mix_incubate_off_time = mix_incubate_off_time
        self._postspin_incubation_time = postspin_incubation_time
        self._remove_wash_vol = remove_wash_vol
        self._sample_mix_height = sample_mix_height
        self._sample_mix_times = sample_mix_times
        self._sample_mix_vol = sample_mix_vol
        self._thermomixer_incubation_time = thermomixer_incubation_time
    
    @labware_loader(5, "_flatplate")
    def load_flatplate(self):
        self._flatplate = self._ctx.load_labware('opentrons_96_aluminumblock_nest_wellplate_100ul', self._flatplate_slot, 'chilled elution plate on block for Station C')
    
    @labware_loader(5, "_tempplate")
    def load_tempplate(self):
        self._tempplate = self._tempdeck.load_labware(self._magplate_model)
    
    @property
    def pcr_samples_m(self):
        return self._flatplate.rows()[0][:self.num_cols]
    
    @property
    def temp_samples_m(self):
        return self._tempplate.rows()[0][:self.num_cols]
    
    def load_etoh(self): pass
    
    @property
    def wash1(self):
        return self._res12.wells()[:6]
    
    @property
    def wash2(self):
        return self._res12.wells()[-6:]
    
    @labware_loader(9, "_elut12")
    def load_elut12(self):
        self._elut12 = self._ctx.load_labware('nest_12_reservoir_15ml', '2', 'Trough with Elution')
    
    @property
    def water(self):
        return self._elut12.wells()[11]
    
    @staticmethod
    def wash_getcol(sample_col_idx: int, wash_cols: int, source):
        return source[sample_col_idx // 2]
    
    def mix_samples(self):
        self._m300.flow_rate.aspirate = 94
        for i, m in enumerate(self.mag_samples_m):
            if self.run_stage("mix sample {}/{}".format(i + 1, len(self.mag_samples_m))):
                self.pick_up(self._m300)
                self._m300.mix(self._sample_mix_times, self._sample_mix_vol, m.bottom(self._sample_mix_height))
                self._m300.air_gap(self._bind_air_gap)
                self.drop(self._m300)
    
    def elute(self, positions=None, transfer: bool = False, stage: str = "elute"):
        if positions is None:
            positions = self.temp_samples_m
        self._magdeck.disengage()
        super(StationBTechnogenetics, self).elute(positions=positions, transfer=transfer, stage=stage)
        self._magdeck.disengage()
    
    def remove_wash(self, vol, stage: str = "remove wash"):
        self.remove_supernatant(vol, stage)
    
    def final_transfer(self):
        self._m300.flow_rate.aspirate = self._final_transfer_rate_aspirate
        self._m300.flow_rate.dispense = self._final_transfer_rate_dispense
        n = len(list(zip(self.mag_samples_m, self.pcr_samples_m)))

        for i, (m, e) in enumerate(zip(self.mag_samples_m, self.pcr_samples_m)):
            if self.run_stage("final transfer {}/{}".format(i + 1, n)):
                self.pick_up(self._m300)
                side = -1 if i % 2 == 0 else 1
                loc = m.bottom(0.3).move(Point(x=side*2))
                self._m300.aspirate(self._final_vol, loc)
                self._m300.air_gap(self._elute_air_gap)
                self._m300.dispense(self._elute_air_gap, e.top())
                self._m300.dispense(self._m300.current_volume, e.bottom(0.5))

                #self._m300.transfer(self._final_vol, loc, e.bottom(self._elution_height), air_gap=self._elute_air_gap, new_tip='never')
                self._m300.mix(self._final_mix_times, self._final_mix_vol, e.bottom(self._final_mix_height))
                self._m300.air_gap(self._elute_air_gap)
                self.drop(self._m300)
    
    def body(self):
        self.logger.info(self.get_msg_format("volume", "wash 1", self._wash_headroom * self._wash_1_vol * self._num_samples / 1000))
        self.logger.info(self.get_msg_format("volume", "wash 2", self._wash_headroom * self._wash_2_vol * self._num_samples / 1000))
        self.logger.info(self.get_msg_format("volume", "elution buffer", self._wash_headroom * self._elution_vol * self._num_samples / 1000))
        self.mix_samples()
        
        if self.run_stage("mix incubate on"):
            self.delay(self._mix_incubate_on_time, self.get_msg_format("incubate on magdeck", self.get_msg("off")))
        self._magdeck.engage(height=self._magheight)
        self.check()
        if self.run_stage("mix incubate off"):
            self.delay(self._mix_incubate_off_time, self.get_msg_format("incubate on magdeck", self.get_msg("on")))
        
        self.remove_supernatant(self._starting_vol)
        self.wash(self._wash_1_vol, self.wash1, self._wash_1_times, "wash A")

        if self.run_stage("spin deepwell wash A"):
            self._magdeck.disengage()
            self.dual_pause("spin the deepwell", between=self.set_external)
            self.set_internal()
            self._magdeck.engage(height=self._magheight)
            self.check()

        if self.run_stage("post spin incubation wash A"):
            self.delay(self._postspin_incubation_time, self.get_msg_format("incubate on magdeck", self.get_msg("on")))

        self.remove_wash(self._remove_wash_vol, "remove wash A after spin")

        if self.run_stage("remove wash A"):
            self.dual_pause("Check Wash A removal and empty waste reservoir")

        self.wash(self._wash_2_vol, self.wash2, self._wash_2_times, "wash B")
        
        if self.run_stage("spin deepwell wash B"):
            self._magdeck.disengage()
            self.dual_pause("spin the deepwell", between=self.set_external)
            self.set_internal()
            self._magdeck.engage(height=self._magheight)
            self.check()
        
        if self.run_stage("post spin incubation wash B"):
            self.delay(self._postspin_incubation_time, self.get_msg_format("incubate on magdeck", self.get_msg("on")))
        
        self.remove_wash(self._remove_wash_vol, "remove wash B after spin")

        if self.run_stage("remove wash B"):
            self.dual_pause("Check Wash B removal and empty waste reservoir")
        
        if self.run_stage("deepwell incubation"):
            self.dual_pause("deepwell incubation", between=self.set_external if self._external_deepwell_incubation else None)
            self.set_internal()
        
        self.elute()
        
        if self.run_stage("thermomixer"):
            self.dual_pause("seal the deepwell", between=self.set_external)
            self.set_internal()
        
        self._magdeck.engage(height=self._magheight)
        self.check()
        if self.run_stage("post thermomixer incubation"):
            self.delay(self._thermomixer_incubation_time, self.get_msg_format("incubate on magdeck", self.get_msg("on")))
        
        if self.run_stage("input PCR"):
            self.dual_pause("input PCR")
        
        self.final_transfer()
        
        self._magdeck.disengage()
        self.logger.info(self.msg_format("move to PCR"))


if __name__ == "__main__":
    StationBTechnogenetics(metadata={'apiLevel': '2.3'}).simulate()


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



import math


class LysisTube:
    """Geometrical model of the Lysis Buffer tube"""
    def __init__(self, radius: float, cone_height: float = 0, fill: float = 0):
        self.radius = radius
        self._ch = cone_height
        self.volume = fill

    @property
    def height(self) -> float:
        r"""Height from inverse formula of setter

        H > h => V > pi * r^2 * h / 3
        (H - h) + h^3 / (3h^2) = V / (pi * r^2) => H = V / (pi * r^2) + 2/3 * h

        H < h => V < pi * r^2 * h / 3
        0 + H^3 / (3h^2) = V / (pi * r^2) => H = cbrt(3h^2 * V / (pi * r^2))
        """
        if self.volume > math.pi * self.radius**2 * self._ch / 3:
            # Filled more than the cone
            return self.volume / (math.pi * self.radius**2) + 2 * self._ch / 3
        else:
            # Filled less than the entire cone
            return (3 * self._ch**2 * self.volume / (math.pi * self.radius**2))**(1/3)

    @height.setter
    def height(self, value: float):
        r"""Volume is the cone volume plus the cylinder volume

        V = max(v - h, 0) * pi * r^2 + min(v, h)^3 * pi / 3 * r^2 / h^2
        V = [max(v - h, 0) + min(v, h)^3 / (3h^2)] * pi * r^2
        """
        self.volume = (
            max(value - self._ch, 0) +                 # cylinder starts at the end of the cone
            min(value, self._ch)**3 / (3*self._ch**2)  # cone ends at self._ch height
        ) * math.pi * self.radius**2

    def extract(self, volume: float) -> float:
        self.volume -= volume
        return self.height

    def fill(self, volume: float):
        self.volume += volume

    def refill(self, volume: float):
        self.volume = volume

    def __str__(self) -> str:
        return "{} with cone radius {:.2f} and height {:.2f} (mm)".format(type(self).__name__, self.radius, self._ch)


# Copyright (c) 2020 Covmatic.
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
